{$A+,B-,E-,F-,G+,I-,N-,O-,V-,X+ }

{$IFDEF DEBUG }
  {$R+,S+ }
{$ELSE }
  {$D-,L-,R-,S- }
{$ENDIF }
{$M 4096, 0, 655360 }


program Simul4;

uses Dos,Crt;

{ *** XMS Error Codes *** }

const Err_NOTIMPLEMENTED = $80;
      Err_VDISKFOUND     = $81;
      Err_A20            = $82;
      Err_GENERAL        = $8E;
      Err_UNRECOVERABLE  = $8F;

      Err_HMANOTEXIST    = $90;
      Err_HMAINUSE       = $91;
      Err_HMAMINSIZE     = $92;
      Err_HMANOTALLOCED  = $93;
      Err_A20ENABLED     = $94;

      Err_OUTOMEMORY     = $A0;
      Err_OUTOHANDLES    = $A1;
      Err_INVALIDHANDLE  = $A2;
      Err_SHINVALID      = $A3;
      Err_SOINVALID      = $A4;
      Err_DHINVALID      = $A5;
      Err_DOINVALID      = $A6;
      Err_LENINVALID     = $A7;
      Err_OVERLAP        = $A8;
      Err_PARITY         = $A9;
      Err_EMBUNLOCKED    = $AA;
      Err_EMBLOCKED      = $AB;
      Err_LOCKOVERFLOW   = $AC;
      Err_LOCKFAIL       = $AD;

      Err_UMBSIZETOOBIG  = $B0;
      Err_NOUMBS         = $B1;
      Err_INVALIDUMB     = $B2;


const XMM_Control : pointer = nil;
      XMM_Error   : byte    = 0;

function XMM_Installed : boolean; assembler;
{ Verifie l'existence du driver XMS et
  note l'adresse du contoleur API }
asm
  mov  ax, 4300h
  int  2Fh
  cmp  al, 80h
  je   @DriverExist
  { pas de drivers }
  xor  al, al    { FALSE }
  jmp  @Exit
@DriverExist:
  { chope l'adresse du controleur }
  mov  ax, 4310h
  int  2Fh
  mov  word ptr XMM_Control, bx
  mov  word ptr XMM_Control+2, es
  xor  al, al
  mov  XMM_Error, al
  mov  al, TRUE
@Exit:
end;


function XMM_Result : word; assembler;
{ renvoi la derniere erreur produite }
asm
  xor  ax, ax
  xchg al, XMM_Error
end;


function XMM_Version : longint; assembler;
{ renvoi la version du driver + version de revision (poids fort) }
asm
  xor  ah, ah
  call XMM_Control
  mov  dx, bx
end;


function XMM_RequestHMA( Space_Needed : word ) : boolean; assembler;
{ reserve de la HMA pour l'appelant }
asm
  mov  ah, 01h
  mov  dx, Space_Needed
  call XMM_Control
  { traite l'erreur eventuelle }
  or   ax, ax
  jnz  @Success
  mov  XMM_Error, bl
@Success:
end;


function XMM_ReleaseHMA : boolean; assembler;
{ libre la HMA reserv prcdement }
asm
  mov  ah, 02h
  call XMM_Control
  { traite l'erreur eventuelle }
  or   ax, ax
  jnz  @Success
  mov  XMM_Error, bl
@Success:
end;


function XMM_GlobalEnableA20 : boolean; assembler;
{ rend disponible la ligne A20 : (rserv  l'admistrateur!) }
asm
  mov  ah, 03h
  call XMM_Control
  or   ax, ax
  jnz  @Success
  mov  XMM_Error, bl
@Success:
end;


function XMM_GlobalDisableA20 : boolean; assembler;
{ rend indisponible la ligne A20 }
asm
  mov  ah, 04h
  call XMM_Control
  or   ax, ax
  jnz  @Success
  mov  XMM_Error, bl
@Success:
end;


function XMM_LocalEnableA20 : boolean; assembler;
{ rend disponible la ligne A20, pour un acces direct en }
{ memoire tendu }
asm
  mov  ah, 05h
  call XMM_Control
  or   ax, ax
  jnz  @Success
  mov  XMM_Error, bl
@Success:
end;


function XMM_LocalDisableA20 : boolean; assembler;
{ rend indiponible la ligne A20 }
asm
  mov  ah, 06h
  call XMM_Control
  or   ax, ax
  jnz  @Success
  mov  XMM_Error, bl
@Success:
end;


function XMM_QueryA20( var A20Enabled : boolean ) : boolean; assembler;
{ cette fonction renvoi TRUE si la ligne A20 est disponible }
asm
  mov  ah, 07h
  call XMM_Control
  or   bl, bl
  jz   @Ok
  mov  XMM_Error, bl
  xor  al, al
  jmp  @Exit
@Ok:
  les  di, A20Enabled
  seges mov [di], al
  mov  al, TRUE
@Exit:
end;


function XMM_QueryFreeExtended( var LargestFree, TotalAmount : word ) : boolean; assembler;
{ renvoi la taille de memoire etendue la plus large disponible, et le
 total disponible en Ko }
asm
  mov  ah, 08h
  call XMM_Control
  or   ax, ax
  jnz  @Ok
  { plus de mmoire }
  mov  XMM_Error, bl
  jmp  @Exit
@Ok:
  les  di, LargestFree
  seges mov [di], ax
  les  di, TotalAmount
  seges mov [di], dx
  mov  al, TRUE
@Exit:
end;


function XMM_AllocateExtended( var Handle : word; Amount : word ) : boolean; assembler;
{ alloue en Ko de la mmoire etendue et renvoi un handle }
asm
  mov  ah, 09h
  mov  dx, Amount
  call XMM_Control
  or   ax, ax
  jz   @Error
  or   dx, dx
  jnz  @Ok
@Error:
  mov  XMM_Error, bl
  xor  al, al
  jmp  @Exit
@Ok:
  les  di, Handle
  seges mov [di], dx    { renvoi le handle }
@Exit:
end;


function XMM_FreeExtended( Handle : word ) : boolean; assembler;
{ libere la memoire etendue precedement allou }
asm
  mov  ah, 0Ah
  mov  dx, Handle
  call XMM_Control
  or   ax, ax
  jnz  @Success
  mov  XMM_Error, bl
@Success:
end;


type ExtMemMoveRec = record
         Length       : longint;  { nombre d'octets  transferer }
         SourceHandle : word;     { 0=>memoire 8086 avec seg:ofs }
         SourceOffset : longint;  { ou seg:ofs si 8086 }
         DestHandle   : word;     { idem... si 0=>... }
         DestOffset   : longint;  { idem... }
       end;

     PExtMemMoveRec = ^ExtMemMoveRec;

function XMM_MoveExtended( var EMMS : ExtMemMoveRec ) : boolean; assembler;
{ transfer de memoire universel }
asm
  mov  ax, SEG XMM_Control
  mov  es, ax
  mov  bx, OFFSET XMM_Control
  push ds       { sauve }
  mov  ah, 0Bh
  lds  si, EMMS
  seges call dword [bx]
  pop  ds       { restore }
  or   ax, ax
  jnz  @Success
  mov  XMM_Error, bl
@Success:
end;


function XMM_LockExtended( Handle : word; var Address : longint ) : boolean; assembler;
{ lock un handle et renvoi l'adresse 32 bits de la zone ram etendue }
asm
  mov  ah, 0Ch
  mov  dx, Handle
  call XMM_Control
  or   ax, ax
  jnz  @Success
  mov  XMM_Error, bl
  jmp  @Exit
@Success:
  les  di, Address
  seges mov [di], bx
  seges mov [di+2], dx
@Exit:
end;


function XMM_UnLockExtended( Handle : word ) : boolean; assembler;
asm
  mov  ah, 0Dh
  mov  dx, Handle
  call XMM_Control
  or   ax, ax
  jnz  @Success
  mov  XMM_Error, bl
@Success:
end;


function XMM_GetHandleInfo( Handle : word;
                            var LockCount : byte;
                            var FreeEMB   : byte;
                            var Length    : word ) : boolean; assembler;
{ renvoi info sur l'handle }
asm
  mov  ah, 0Eh
  mov  dx, Handle
  call XMM_Control
  or   ax, ax
  jnz  @Success
  mov  XMM_Error, bl
  jmp  @Exit
@Success:
  les  di, LockCount
  seges mov [di], bh
  les  di, FreeEMB
  seges mov [di], bl
  les  di, Length
  seges mov [di], dx
@Exit:
end;


function XMM_ReallocateExtended( Handle, NewSize : word ) : boolean; assembler;
{ modifie la taille d'une allocation }
asm
  mov  ah, 0Fh
  mov  dx, Handle
  mov  bx, NewSize
  call XMM_Control
  or   ax, ax
  jnz  @Success
  mov  XMM_Error, bl
@Success:
end;


function XMM_RequestUMB( Size : word;
                         var SegUMB : word;
                         var SizeUMB : word ) : boolean; assembler;
{ tente d'allouer un block en dessous des 1 Megas et au dessus des 640Ko
  Size est en paragraphe.
  Renvoi FALSE si erreur. si XMM_Result=B0h alors SizeUMB=taille max libre }
asm
  mov  ah, 10h
  mov  dx, Size
  call XMM_Control
  les  di, SizeUMB
  mov  [di], dx      { dans tous les cas, on note DX (SizeUMB) }
  or   ax, ax
  jnz  @Success
  mov  XMM_Error, bl
  jmp  @Exit
@Success:
  les  di, SegUMB
  mov  [di], bx
@Exit:
end;


function XMM_ReleaseUMB( SegUMB : word ) : boolean; assembler;
{ libre un block Upper Memory prcdement allou }
asm
  mov  ah, 11h
  mov  dx, SegUMB
  call XMM_Control
  or   ax, ax
  jnz  @Success
  mov  XMM_Error, bl
@Success:
end;


procedure AffErreur;
begin
  writeln('XMM Error : ',XMM_Result);
end;


function FoncTas( Taille : word ) : integer; far;
begin
  if Taille>0 then FoncTas := 1;
end;


{ *** Task *** }

const MemHeadMark = $E2E4; { marque d'integrit de struc. (si alloue) }
      MemFreeMark = $A1A2; { marque d'integrit de struc. (si libre) }
      EndLcMem    = $FFFF; { decalage sans suite, pour LcMem }

type PLcMemRec = ^LcMemRec;

     LcMemRec = record
         HeadMark  : word;     { marque d'integrit de structure }
         Level     : longint;  { nd'ordre d'entre }
         Prec      : word;     { offset relatif au PF donnes (EndLcMem=>rien) }
         Next      : word;     { idem (Prec<Next) }
         Lg        : word;     { longueur espace mmoire }
         CheckSum  : word;     { pour le mode DEBUG }
         EndMark   : word;     { marque d'integrit de structure }
       end;

     TaskMode = (TaskOFF,TaskStandBY,TaskON);

     TaskRec = record
         St         : TaskMode;

         SS         : word;
         SP         : word;
         BP         : word;

         LcMemAvail : word;     { mmoire libre (max:28 Ko) }
         FirstLcMem : word;     { premiere allocation }
         LastLcMem  : word;     { dernire ... }
         NbLcMem    : word;     { nombre d'allocation }
       end;

     PTaskRec = ^TaskRec;


const LgTabTask = 32;

var TabTask : array[0..LgTabTask-1] of TaskRec;

    NbTask     : word;        { nombre total de tches }
    NumTask    : word;        { n de la tche courante }
    _Task_     : PTaskRec;    { structure de la tche courante }
    NbTaskON   : word;        { nombre de tches actives }

    LgStack    : word;        { taille de la pile d'une tche }
    LcGranule  : word;        { Taux de granularit: 0,1,3,... }

    VARSS      : word;
    VARSP      : word;
    VARBP      : word;
    VAREMM     : ExtMemMoveRec;

    BasicSS    : word;        { etat de la pile orginel }
    BasicSP    : word;
    BasicBP    : word;


var XMSHandle   : word;        { Handle XMS en fonction de la mem alloue }
{    SegSys      : word; }       { Segment rserv  Zedix }
{    SegPg       : word; }       { Segment rserv aux pages locales }
    SegLcMem    : word;        { Segment rserv aux donnes locales }
    SegStack    : word;        { SS de la pile tche }
    SegSwap     : word;        { Segment de base contenant SegPg+LcMem+Stack }
    SwapSize    : word;        { longueur total (max:65535) }



procedure ChangeTask; forward;  { ---- voir plus loin }


{ --------------------------------
  Messages affichs par les tches
  -------------------------------- }
procedure MsgTask( texte : string );
begin
  writeln('[',NumTask:2,']: ',texte);
end;


{ ----------------------------------
  Renvoi la mmoire local disponible
  ---------------------------------- }
function ixLcMemAvail : word;
begin
  ixLcMemAvail := _Task_^.LcMemAvail;
end;



{ -------------------------------------------
  Recherche en LcMem une partie libre
  ------------------------------------------- }
procedure GetLcMem( var Ofs : word;         { position debut libre }
                    var BA  : PLcMemRec; { structure suivant la pos. (nil=>fin) }
                    size    : word );       { longueur minimun recherch }
{ ----------------------------------
  Ofs=EndLcMem si pas assez de place
  ---------------------------------- }
type PtrRec = record Ofs, Seg : word; end;
var g : word;
    d : word;
    h : PLcMemRec;
begin
  d := 0;
  PtrRec(h).Seg := SegLcMem;
  g := _Task_^.FirstLcMem;

  while g<>EndLcMem do
    begin
      PtrRec(h).Ofs := g;
      if g-d>=size then
        begin
          { ---------------------
            zone libre suffisante
            --------------------- }
          if (h^.HeadMark=MemHeadMark) and
             (h^.EndMark=MemHeadMark) then
            begin
              Ofs := d;
              BA  := h;
            end
          else
            begin
              Ofs := EndLcMem;
              BA  := h; { il y a erreur=>plus de mem }
            end;
          exit;
        end;
      d := g+sizeof(LcMemRec)+h^.Lg;
      g := h^.Next;
    end;

  if (32768-LgStack)-d>=size then
    begin
      Ofs := d;
      BA  := nil;    { nil => la derniere structure }
    end
  else
    begin
      Ofs := EndLcMem; { rien }
      BA  := nil;         { absolument rien }
    end;
end;


{ -------------------------------
  Alloue un bloc de mmoire local
  ------------------------------- }
function ixLcAlloc( var p : pointer; size : word ) : boolean;
type PtrRec = record Ofs, Seg : word; end;
var a : pointer;
    x : longint;
    o : word;
    h : PLcMemRec;
    l : PLcMemRec;
    n : PLcMemRec;
begin
  size := (size+LcGranule) and not(LcGranule);
  if (size>0) and (size<$FFFF-sizeof(LcMemRec)) and
     (_Task_^.LcMemAvail>=sizeof(LcMemRec)+size) then
    begin
      GetLcMem(o,h,size+sizeof(LcMemRec));
      if o<>EndLcMem then
        begin
          n := ptr(SegLcMem,o);
          FillChar(n^,sizeof(n^),0);
          n^.HeadMark := MemHeadMark;
          n^.Level    := 0;
          if h=nil then
            begin
              if _Task_^.LastLcMem=EndLcMem then
                begin
                  _Task_^.FirstLcMem := o;
                  n^.Prec := EndLcMem;
                end
              else
                begin
                  l := ptr(SegLcMem,_Task_^.LastLcMem);
                  l^.Next := o;
                  n^.Prec := PtrRec(l).Ofs;
                end;
              _Task_^.LastLcMem := o;
              n^.Next := EndLcMem;
            end
          else
            begin
              if h^.Prec=EndLcMem then
                begin
                  _Task_^.FirstLcMem := o;
                  n^.Prec := EndLcMem;
                end
              else
                begin
                  l := ptr(SegLcMem,h^.Prec);
                  l^.Next := o;
                  n^.Prec := PtrRec(l).Ofs;
                end;
              h^.Prec := o;
              n^.Next := PtrRec(h).Ofs;
            end;
          n^.Lg       := size;
          n^.CheckSum := 0;
          n^.EndMark  := MemHeadMark;

          inc(_Task_^.NbLcMem);
          dec(_Task_^.LcMemAvail,size+sizeof(LcMemRec));

          x := (longint(PtrRec(n).Seg) shl 4)+
               longint(PtrRec(n).Ofs)+sizeof(LcMemRec);
                 { --- adresse reel }
          PtrRec(a).Seg := word(x shr 4);
          PtrRec(a).Ofs := word(x) and 15;  { --- adresse standardis }
        end
      else a := nil; { plus ou pas assez de mmoire }
    end
  else a := nil;
  p := a;
  ixLcAlloc := a<>nil;
end;



{ ------------------------
  Libere la mmoire locale
  ------------------------ }
procedure ixLcFree( var p : pointer );
{ -----------------------------------
  Le pointer est toujours renvoy NIL
  ----------------------------------- }
type PtrRec = record Ofs, Seg : word; end;
var a : pointer;
    x : longint;
    o : word;
    n : PLcMemRec;
begin
  a := p;
  if a<>nil then
    begin
      x := (longint(PtrRec(a).Seg) shl 4)+
           longint(PtrRec(a).Ofs);             { --- adresse reel }
      PtrRec(a).Seg := word(x shr 4);
      PtrRec(a).Ofs := word(x) and 15;          { --- adresse standardis }

      { ------------------
        Adresse realiste ?
        ------------------ }
      if (PtrRec(a).Seg>=SegLcMem) and
         (PtrRec(a).Seg<SegStack) then
        begin { --- Oui }
          { ----------------------------------------
            position de la structure de base mmoire
            ---------------------------------------- }
          o := ((PtrRec(a).Seg-SegLcMem)*16+PtrRec(a).Ofs)-
               sizeof(LcMemRec);
          n := ptr(SegLcMem,o);
          if (n^.HeadMark=MemHeadMark) and
             (n^.EndMark=MemHeadMark) then
            begin
              n^.HeadMark := MemFreeMark;
              n^.EndMark  := MemFreeMark;
              if n^.Prec<>EndLcMem then
                LcMemRec(ptr(SegLcMem,n^.Prec)^).Next := n^.Next
              else _Task_^.FirstLcMem := n^.Next;
              if n^.Next<>EndLcMem then
                LcMemRec(ptr(SegLcMem,n^.Next)^).Prec := n^.Prec
              else _Task_^.LastLcMem := n^.Prec;
              dec(_Task_^.NbLcMem);
              inc(_Task_^.LcMemAvail,n^.Lg+sizeof(LcMemRec));
            end;
        end;
      p := nil;
    end;
end;



{ ---------------------------------------------------------
  Renvoi le niveau de fragmentation de la mmoire local
  (renvoi toujours "0" si la mmoire local est dans le tas)
  --------------------------------------------------------- }
function ixLcFrag : word;
type PtrRec = record Ofs, Seg : word; end;
var g : word;
    d : word;
    h : PLcMemRec;
    n : word;
begin
  n := 0;
  d := 0;
  PtrRec(h).Seg := SegLcMem;
  g := _Task_^.FirstLcMem;
  while g<>EndLcMem do
    begin
      PtrRec(h).Ofs := g;
      if g-d>0 then inc(n);
      d := g+sizeof(LcMemRec)+h^.Lg;
      g := h^.Next;
    end;
  ixLcFrag := n;
end;



{ --------------------------
  Renvoi un nombre en chaine
  -------------------------- }
function StrVal( x : longint ) : string;
var s : string[20];
begin
  Str(x,s);
  StrVal := s;
end;


procedure _ShowAlloc;
type PtrRec = record Ofs, Seg : word; end;
var h : PLcMemRec;
    g : word;
    d : word;
    l : word;
    c : word;
begin
  TextAttr := $07;
  ClrScr;

  GotoXY(1,1);
  TextAttr := $07;
  writeln('ixLcMemAvail=',ixLcMemAvail:5);
  writeln('    ixLcFrag=',ixLcFrag:5);
  writeln('     NbAlloc=',_Task_^.NbLcMem:5);
  g := _Task_^.LastLcMem;
  if g=$FFFF then writeln
  else
    begin
      h := ptr(SegLcMem,g);
      writeln('         Top=',g+h^.Lg+sizeof(LcMemRec):5);
    end;
  writeln('    SegLcMem=',SegLcMem:5);

  d := 0;
  l := 0;
  c := 18;
  PtrRec(h).Seg := SegLcMem;
  g := _Task_^.FirstLcMem;
  while g<>EndLcMem do
    begin
      PtrRec(h).Ofs := g;
      if g-d>0 then
        begin
          GotoXY(c+1,l+1);
          TextAttr := $0F;
          write('|');
          TextAttr := $07;
          write(g-d:5);
          inc(l);
          if l>24 then
            begin
              l := 0;
              inc(c,6);
            end;
        end;
      GotoXY(c+1,l+1);
      TextAttr := $0F;
      write('|');
      TextAttr := $70;
      write(sizeof(LcMemRec)+h^.Lg:5);
      inc(l);
      if l>24 then
        begin
          l := 0;
          inc(c,6);
        end;

      d := g+sizeof(LcMemRec)+h^.Lg;
      g := h^.Next;
    end;
end;



procedure ShowAlloc;
type PtrRec = record Ofs, Seg : word; end;
var n : PLcMemRec;
    d : word;
begin
  PtrRec(n).Seg := SegLcMem;
  d := _Task_^.FirstLcMem;
  while d<>EndLcMem do
    begin
      PtrRec(n).Ofs := d;
      if (n^.HeadMark=MemHeadMark) and
         (n^.EndMark=MemHeadMark) then
        begin
          MsgTask('Alloc en '+StrVal(d)+' de '+StrVal(n^.Lg)+' octets');
          d := n^.Next;
        end
      else
        begin
          MsgTask('Invalide struc.');
          d := EndLcMem;
        end;
    end;
  MsgTask('--- Fin Alloc. '+StrVal(ixLcMemAvail));
end;



{ ----------
  La tche 0
  ---------- }
procedure ProcessTask0;
var i : word;
begin
  MsgTask('Multitche en cours... '+StrVal(MemAvail)+'/'+StrVal(ixLcMemAvail));

  for i := 1 to 100 do ChangeTask;

  if NbTaskON>1 then
    begin
      MsgTask('Il existe encore des tches en cours...');
      while (NbTaskON>1) and not(KeyPressed) do ChangeTask;
    end;

  MsgTask('Fin de toutes les tches.'+StrVal(ixLcMemAvail)+' Frag='+StrVal(ixLcFrag));
end;



(*
procedure ProcessTask0;

  procedure ShowAlloc;

    procedure WithEms;
    type PtrRec = record Ofs, Seg : word; end;
    var h : PEmsAllocRec;
        g : word;
        d : word;
        l : word;
        c : word;
    begin
      TextAttr := $07;
      ClrScr;

      GotoXY(1,1);
      TextAttr := $07;
      writeln('ixLcMemAvail=',ixLcMemAvail:5);
      writeln('    ixLcFrag=',ixLcFrag:5);
      writeln('     NbAlloc=',_Task_^.Ems.NbAlloc:5);
      g := _Task_^.Ems.LastAlloc;
      if g=$FFFF then writeln
      else
        begin
          h := ptr(EmsDataSeg,g);
          writeln('         Top=',g+h^.Lg+sizeof(EmsAllocRec):5);
        end;
      writeln('  EmsDataSeg=',EmsDataSeg:5);

      d := 0;
      l := 0;
      c := 18;
      PtrRec(h).Seg := EmsDataSeg;
      g := _Task_^.Ems.FirstAlloc;
      while g<>EmsOfsNIL do
        begin
          PtrRec(h).Ofs := g;
          if g-d>0 then
            begin
              GotoXY(c+1,l+1);
              TextAttr := $0F;
              write('|');
              TextAttr := $07;
              write(g-d:5);
              inc(l);
              if l>24 then
                begin
                  l := 0;
                  inc(c,6);
                end;
            end;
          GotoXY(c+1,l+1);
          TextAttr := $0F;
          write('|');
          TextAttr := $70;
          write(sizeof(EmsAllocRec)+h^.Lg:5);
          inc(l);
          if l>24 then
            begin
              l := 0;
              inc(c,6);
            end;

          d := g+sizeof(EmsAllocRec)+h^.Lg;
          g := h^.Next;
        end;
    end;

    procedure WithoutEms;
    begin
    end;

  begin
    if EmsON then WithEms else WithoutEms;
  end;


const LgT = 300;
      MxSize = 256;

var t : array[1..LgT] of pointer;

var i : word;
    nb : word;
    n : word;
    l : word;
    c : char;
    Wait : boolean;
    nbtest : word;
    nbget  : word;
    nbfree : word;
    lg     : word;
begin
  Randomize;

  for i := 1 to LgT do t[i] := nil;

  Wait := FALSE;

  nb := 0;
  nbget := 0;
  nbfree := 0;
  nbtest := 0;
  c := ' ';
  while UpCase(c)<>'Q' do
    begin
      n := Random(LgT)+1;
      if t[n]=nil then
        begin
          lg := Random(MxSize);
          if lg=0 then lg := 1;
          if ixLcAlloc(t[n],lg) then
            begin
              inc(nbget);
              inc(nb);
            end
          else
            begin
              sound(2000);
              delay(200);
              nosound;
              Wait := TRUE;
            end;
        end
      else
        if ((_Task_^.Ems.NbAlloc*100) div LgT)>=0 then
        begin
          inc(nbfree);
          ixLcFree(t[n]);
          t[n] := nil;
          dec(nb);
        end;
      inc(nbtest);
      ShowAlloc;

      if KeyPressed or Wait then
        begin

          {
          GotoXY(1,1);
          writeln('Fragment=',ixLcFrag:10);
          writeln('Nbtest  =',nbtest:10);
          writeln('Nb      =',nb:10);
          writeln('MemAvail=',ixLcMemAvail:10);
          writeln('MaxAvail=',MaxAvail:10);
          writeln('NbGet   =',nbget:10);
          writeln('NbFree  =',nbfree:10);
          for i := 1 to LgT do
            if t[i]<>nil then SetMem(i,1) else SetMem(i,0);
          }
          if KeyPressed then c := ReadKey;
          repeat
          until KeyPressed;
          ClrScr;
          c := ReadKey;
          Wait := FALSE;
        end;
    end;


  MsgTask('Fin de toutes les tches.'+StrVal(ixLcMemAvail)+' Frag='+StrVal(ixLcFrag));

  for i := 1 to LgT do ixLcFree(t[i]);

  MsgTask('Bis: '+StrVal(ixLcMemAvail)+' Frag='+StrVal(ixLcFrag));
end;
*)



{ ---------------
  Les tches (>0)
  --------------- }
procedure ProcessTaskX;
const LgT = 1000;

type TRec = record
         p  : pointer;
         lg : word;
       end;

     Buffer = array[1..10000] of byte;

var i : word;
    t : array[1..LgT] of TRec;
    n : word;
    lg : word;
    j : word;
begin
  Randomize;

  for i := 1 to LgT do
    begin
      t[i].p  := nil;
      t[i].lg := 0;
    end;

  MsgTask('Nouvelle tche... '+StrVal(MemAvail div 1024)+' Ko Free et '+
          StrVal(ixLcMemAvail));
  for i := 1 to 80 do
    begin
      n := Random(LgT)+1;
      if t[n].p=nil then
        begin
          lg := Random(512)+1;
          if ixLcAlloc(t[n].p,lg) then
            begin
              t[n].lg := lg;
              FillChar(t[n].p^,lg,byte(n));
            end;
        end
      else
        begin
          ixLcFree(t[n].p);
          t[n].p  := nil;
          t[n].lg := 0;
        end;
      ChangeTask;
    end;
  MsgTask('Fin de tche. '+StrVal(MemAvail div 1024)+' Ko Free ('+StrVal(ixLcMemAvail)+')');

  for i := 1 to LgT do
    if t[i].p<>nil then
      begin
        for j := 1 to t[i].lg do
          if Buffer(t[i].p^)[j]<>byte(i) then
            begin
              MsgTask('i='+StrVal(i)+' j='+StrVal(j));
              sound(2000);
              delay(200);
              nosound;
              j := t[i].lg;
              i := LgT;
            end;
      end;
end;



{ -------------------------------
  Effectue le changement de tche
  ------------------------------- }
procedure ChangeTask;
begin
  { ---------------------------------
    Sauvegarde le contexte de la pile
    --------------------------------- }
  asm
    mov  VARBP, bp
    mov  VARSP, sp
    mov  ax, ss
    mov  VARSS, ax
  end;
  _Task_^.SS := VARSS;
  _Task_^.SP := VARSP;
  _Task_^.BP := VARBP;

  { --------------------------------------------------
    On se place sur une pile stable (la pile standard)
    -------------------------------------------------- }
  asm
    cli
    mov  bp, BasicBP
    mov  sp, BasicSP
    mov  ax, BasicSS
    mov  ss, ax
    sti
  end;

  { ------------------------------------------
    Transfert les donnes de la tche courante
    ------------------------------------------ }
  VAREMM.Length       := SwapSize;
  VAREMM.SourceHandle := 0;
  VAREMM.SourceOffset := longint(ptr(SegSwap,0));
  VAREMM.DestHandle   := XMSHandle;
  VAREMM.DestOffset   := longint(SwapSize)*longint(NumTask);
  XMM_MoveExtended(VAREMM);  { SegSwap->Extended }

  while TRUE do
    begin
      inc(NumTask);
      if NumTask>=NbTask then NumTask := 0;
      _Task_ := @TabTask[NumTask];
      case _Task_^.St of
        TaskStandBy :
          begin

            inc(NbTaskON);

            { -------------------------------
              Activation d'une nouvelle tche
              ------------------------------- }
            FillChar(_Task_^,sizeof(_Task_^),0);
            _Task_^.St := TaskON;
            _Task_^.SS := SegStack;

            _Task_^.FirstLcMem := EndLcMem;
            _Task_^.LastLcMem  := EndLcMem;
            _Task_^.LcMemAvail := 32768-LgStack;

            _Task_^.SP := LgStack;
            _Task_^.BP := LgStack;

            { -------------
              Nouvelle pile
              ------------- }
            VARSS := _Task_^.SS;
            VARSP := _Task_^.SP;
            VARBP := _Task_^.BP;

            asm
              cli
              mov  bp, VARBP
              mov  sp, VARSP
              mov  ax, VARSS
              mov  ss, ax
              sti
            end;

            ProcessTaskX;

            { --------------------------------------------------
              On se place sur une pile stable (la pile standard)
              -------------------------------------------------- }
            asm
              cli
              mov  bp, BasicBP
              mov  sp, BasicSP
              mov  ax, BasicSS
              mov  ss, ax
              sti
            end;

            _Task_^.St := TaskOFF;
            dec(NbTaskON);
          end;
        TaskON :
          begin
            { ------------------
              Charge les donnees
              ------------------ }
            VAREMM.Length       := SwapSize;
            VAREMM.SourceHandle := XMSHandle;
            VAREMM.SourceOffset := longint(SwapSize)*longint(NumTask);
            VAREMM.DestHandle   := 0;
            VAREMM.DestOffset   := longint(ptr(SegSwap,0));
            XMM_MoveExtended(VAREMM);

            { ------------
              Fixe la pile
              ------------ }
            VARSS := _Task_^.SS;
            VARSP := _Task_^.SP;
            VARBP := _Task_^.BP;
            asm
              cli
              mov  bp, VARBP
              mov  sp, VARSP
              mov  ax, VARSS
              mov  ss, ax
              sti
            end;

            exit;    { --- et on repart }
          end;
        TaskOFF : ;  { --- rien }
      end;
    end;
end;



{ ----------------------------
  Initialisation de la tche 0
  ---------------------------- }
procedure InitTask0;
begin
  { -------------------------------------
    On conserve l'etat courant de la pile
    (on s'en servira plus tard)
    ------------------------------------- }
  asm
    mov  BasicBP, bp
    mov  BasicSP, sp
    mov  ax, ss
    mov  BasicSS, ax
  end;

  { -----------------------------------
    Initialise la tche 0: Pile, etc...
    ----------------------------------- }
  NbTaskON   := 1;
  NumTask    := 0;
  _Task_     := @TabTask[0];
  FillChar(_Task_^,sizeof(_Task_^),0);
  _Task_^.St := TaskON;
  _Task_^.FirstLcMem := EndLcMem;
  _Task_^.LastLcMem  := EndLcMem;
  _Task_^.LcMemAvail := 32768-LgStack;
  _Task_^.SS := SegStack;

  _Task_^.SP := LgStack;
  _Task_^.BP := LgStack;

  { -------------
    Nouvelle pile
    ------------- }
  VARSS := _Task_^.SS;
  VARSP := _Task_^.SP;
  VARBP := _Task_^.BP;
  asm
    cli
    mov  bp, VARBP
    mov  sp, VARSP
    mov  ax, VARSS
    mov  ss, ax
    sti
  end;

  ProcessTask0;    { --- depart de la tche 0 }

  { ------------------------------------
    On restaure l'etat normal de la pile
    ------------------------------------ }
  asm
    cli
    mov  bp, BasicBP
    mov  sp, BasicSP
    mov  ax, BasicSS
    mov  ss, ax
    sti
  end;
end;


var SaveExit : pointer;


procedure FreeXMS; far;
begin
  ExitProc := SaveExit;
  { -------------------------
    Libere la memoire alloue
    ------------------------- }
  XMM_FreeExtended(XMSHandle);
  writeln('XMS libre');
end;



function XMSOpen : boolean;
type LongRec = record Low, High : word; end;
var A20En : boolean;
    LFree, TFree : word;
    MinKo : word;
    ver   : longint;
begin
  XMSOpen := FALSE;
  if XMM_Installed then
    begin
      ver := XMM_Version;
      writeln('XMM Version : ',Hi(LongRec(ver).Low),'.',
                               Lo(LongRec(ver).Low),'  (Rvision:',
                               Hi(LongRec(ver).High),'.',
                               Lo(LongRec(ver).High),')');
      if Hi(LongRec(ver).Low)>=2 then
        if XMM_QueryA20( A20En ) then
          if not(A20En) then
            if XMM_QueryFreeExtended(LFree,TFree) then
              begin
                MinKo := 32*NbTask;

                if MinKo<=LFree then
                  if XMM_AllocateExtended(XMSHandle,MinKo) then
                    begin
                      SaveExit := ExitProc;
                      ExitProc := @FreeXMS;

                      SegLcMem := SegSwap;
                      SegStack := SegSwap+2048-(LgStack div 16);
                      XMSOpen  := TRUE;
                    end
                  else AffErreur
                else writeln('Pas assez de mmoire: ',LFree,'<',MinKo);
              end
            else AffErreur
          else writeln('Registre A20 inutilisable')
        else AffErreur
      else writeln('Version 2.00 ou suprieur seulement');
    end
  else writeln('Pas de driver XMS (HIMEM.SYS)');
end;


procedure XMSClose;
begin
  ExitProc := SaveExit;
  { -------------------------
    Libere la memoire alloue
    ------------------------- }
  XMM_FreeExtended(XMSHandle);
  writeln('XMS libre');
end;



{ -------------------------
  Initialise l'environement
  ------------------------- }
procedure LaunchSys;
type SwapAreaDef = array[1..32768+16] of byte;
     PtrRec = record Ofs, Seg : word; end;
var SwapArea : ^SwapAreaDef;
    i : word;
begin
  { -----------------------------
    Correction du nombre de tche
    ----------------------------- }
  if NbTask<=0 then NbTask := 1
  else
    if NbTask>LgTabTask then NbTask := LgTabTask;

  { -------------------------------------
    Correction sur la longueur de la pile
    ------------------------------------- }
  LgStack := (LgStack+15) and not(15);
  if LgStack<4096 then LgStack := 4096      { --- 4 Ko minimun }
  else
    if LgStack>28672 then LgStack := 28672;  { --- 28 Ko maximun }

  { -------------------------------------------------
    Creation d'une zone "Swap" entre Conv et Extended
    ------------------------------------------------- }
  New(SwapArea);
  if SwapArea<>nil then
    begin
      if PtrRec(SwapArea).Ofs<>0 then SegSwap := PtrRec(SwapArea).Seg+1
      else SegSwap := PtrRec(SwapArea).Seg;
      SwapSize := 32768;

      { ----------------------
        Initialisation mmoire
             (ou non)
        ---------------------- }
      if XMSOpen then
        begin
          { --------------------------------------
            Initialisation des tches potentielles
            -------------------------------------- }
          for i := 0 to NbTask-1 do
            begin
              FillChar(TabTask[i],sizeof(TabTask[i]),0);
              TabTask[i].St := TaskOFF;
            end;

          { ----------------------------------------------
            Mettre quelques tches en attente de lancement
            ---------------------------------------------- }

          TabTask[1].St := TaskStandBy;
          TabTask[5].St := TaskStandBy;
          TabTask[7].St := TaskStandBy;
          TabTask[8].St := TaskStandBy;


          InitTask0;

          XMSClose;  { --- Liberer la mmoire }
        end;
      Dispose(SwapArea);
    end;
end;


begin
  HeapError := @FoncTas;

  NbTask    := 10;
  LgStack   := 8192;
  LcGranule := (1 shl 4)-1;   { 0=>0, 1=>1, 2=>3, 3=>7, 4=>15 ... 7=>127 }

  writeln('Tasks Simulator IV');
  LaunchSys;
end.
