{$A+,B-,E-,F-,G+,I-,N-,O-,V-,X+ }

{$IFDEF DEBUG }
  {$R+,S+ }
{$ELSE }
  {$D-,L-,R-,S- }
{$ENDIF }
{$M 16384, 0, 655360 }

program XMStest;

{ *** XMS Error Codes *** }

const Err_NOTIMPLEMENTED = $80;
      Err_VDISKFOUND     = $81;
      Err_A20            = $82;
      Err_GENERAL        = $8E;
      Err_UNRECOVERABLE  = $8F;

      Err_HMANOTEXIST    = $90;
      Err_HMAINUSE       = $91;
      Err_HMAMINSIZE     = $92;
      Err_HMANOTALLOCED  = $93;
      Err_A20ENABLED     = $94;

      Err_OUTOMEMORY     = $A0;
      Err_OUTOHANDLES    = $A1;
      Err_INVALIDHANDLE  = $A2;
      Err_SHINVALID      = $A3;
      Err_SOINVALID      = $A4;
      Err_DHINVALID      = $A5;
      Err_DOINVALID      = $A6;
      Err_LENINVALID     = $A7;
      Err_OVERLAP        = $A8;
      Err_PARITY         = $A9;
      Err_EMBUNLOCKED    = $AA;
      Err_EMBLOCKED      = $AB;
      Err_LOCKOVERFLOW   = $AC;
      Err_LOCKFAIL       = $AD;

      Err_UMBSIZETOOBIG  = $B0;
      Err_NOUMBS         = $B1;
      Err_INVALIDUMB     = $B2;


const XMM_Control : pointer = nil;
      XMM_Error   : byte    = 0;

function XMM_Installed : boolean; assembler;
{ Verifie l'existence du driver XMS et
  note l'adresse du contoleur API }
asm
  mov  ax, 4300h
  int  2Fh
  cmp  al, 80h
  je   @DriverExist
  { pas de drivers }
  xor  al, al    { FALSE }
  jmp  @Exit
@DriverExist:
  { chope l'adresse du controleur }
  mov  ax, 4310h
  int  2Fh
  mov  word ptr XMM_Control, bx
  mov  word ptr XMM_Control+2, es
  xor  al, al
  mov  XMM_Error, al
  mov  al, TRUE
@Exit:
end;


function XMM_Result : word; assembler;
{ renvoi la derniere erreur produite }
asm
  xor  ax, ax
  xchg al, XMM_Error
end;


function XMM_Version : longint; assembler;
{ renvoi la version du driver + version de revision (poids fort) }
asm
  xor  ah, ah
  call XMM_Control
  mov  dx, bx
end;


function XMM_RequestHMA( Space_Needed : word ) : boolean; assembler;
{ reserve de la HMA pour l'appelant }
asm
  mov  ah, 01h
  mov  dx, Space_Needed
  call XMM_Control
  { traite l'erreur eventuelle }
  or   ax, ax
  jnz  @Success
  mov  XMM_Error, bl
@Success:
end;


function XMM_ReleaseHMA : boolean; assembler;
{ libre la HMA reserv prcdement }
asm
  mov  ah, 02h
  call XMM_Control
  { traite l'erreur eventuelle }
  or   ax, ax
  jnz  @Success
  mov  XMM_Error, bl
@Success:
end;


function XMM_GlobalEnableA20 : boolean; assembler;
{ rend disponible la ligne A20 : (rserv  l'admistrateur!) }
asm
  mov  ah, 03h
  call XMM_Control
  or   ax, ax
  jnz  @Success
  mov  XMM_Error, bl
@Success:
end;


function XMM_GlobalDisableA20 : boolean; assembler;
{ rend indisponible la ligne A20 }
asm
  mov  ah, 04h
  call XMM_Control
  or   ax, ax
  jnz  @Success
  mov  XMM_Error, bl
@Success:
end;


function XMM_LocalEnableA20 : boolean; assembler;
{ rend disponible la ligne A20, pour un acces direct en }
{ memoire tendu }
asm
  mov  ah, 05h
  call XMM_Control
  or   ax, ax
  jnz  @Success
  mov  XMM_Error, bl
@Success:
end;


function XMM_LocalDisableA20 : boolean; assembler;
{ rend indiponible la ligne A20 }
asm
  mov  ah, 06h
  call XMM_Control
  or   ax, ax
  jnz  @Success
  mov  XMM_Error, bl
@Success:
end;


function XMM_QueryA20( var A20Enabled : boolean ) : boolean; assembler;
{ cette fonction renvoi TRUE si la ligne A20 est disponible }
asm
  mov  ah, 07h
  call XMM_Control
  or   bl, bl
  jz   @Ok
  mov  XMM_Error, bl
  xor  al, al
  jmp  @Exit
@Ok:
  les  di, A20Enabled
  mov  [es:di], al
  mov  al, TRUE
@Exit:
end;


function XMM_QueryFreeExtended( var LargestFree, TotalAmount : word ) : boolean; assembler;
{ renvoi la taille de memoire etendue la plus large disponible, et le
 total disponible en Ko }
asm
  mov  ah, 08h
  call XMM_Control
  or   ax, ax
  jnz  @Ok
  { plus de mmoire }
  mov  XMM_Error, bl
  jmp  @Exit
@Ok:
  les  di, LargestFree
  mov  [es:di], ax
  les  di, TotalAmount
  mov  [es:di], dx
  mov  al, TRUE
@Exit:
end;


function XMM_AllocateExtended( var Handle : word; Amount : word ) : boolean; assembler;
{ alloue en Ko de la mmoire etendue et renvoi un handle }
asm
  mov  ah, 09h
  mov  dx, Amount
  call XMM_Control
  or   ax, ax
  jz   @Error
  or   dx, dx
  jnz  @Ok
@Error:
  mov  XMM_Error, bl
  xor  al, al
  jmp  @Exit
@Ok:
  les  di, Handle
  mov  [es:di], dx    { renvoi le handle }
@Exit:
end;


function XMM_FreeExtended( Handle : word ) : boolean; assembler;
{ libere la memoire etendue precedement allou }
asm
  mov  ah, 0Ah
  mov  dx, Handle
  call XMM_Control
  or   ax, ax
  jnz  @Success
  mov  XMM_Error, bl
@Success:
end;


type ExtMemMoveRec = record
         Length       : longint;  { nombre d'octets  transferer }
         SourceHandle : word;     { 0=>memoire 8086 avec seg:ofs }
         SourceOffset : longint;  { ou seg:ofs si 8086 }
         DestHandle   : word;     { idem... si 0=>... }
         DestOffset   : longint;  { idem... }
       end;

     PExtMemMoveRec = ^ExtMemMoveRec;

function XMM_MoveExtended( var EMMS : ExtMemMoveRec ) : boolean; assembler;
{ transfer de memoire universel }
asm
  mov  ax, SEG XMM_Control
  mov  es, ax
  mov  bx, OFFSET XMM_Control
  push ds       { sauve }
  mov  ah, 0Bh
  lds  si, EMMS
  seges call dword [bx]
  pop  ds       { restore }
  or   ax, ax
  jnz  @Success
  mov  XMM_Error, bl
@Success:
end;


function XMM_LockExtended( Handle : word; var Address : longint ) : boolean; assembler;
{ lock un handle et renvoi l'adresse 32 bits de la zone ram etendue }
asm
  mov  ah, 0Ch
  mov  dx, Handle
  call XMM_Control
  or   ax, ax
  jnz  @Success
  mov  XMM_Error, bl
  jmp  @Exit
@Success:
  les  di, Address
  mov  [es:di], bx
  mov  [es:di+2], dx
@Exit:
end;


function XMM_UnLockExtended( Handle : word ) : boolean; assembler;
asm
  mov  ah, 0Dh
  mov  dx, Handle
  call XMM_Control
  or   ax, ax
  jnz  @Success
  mov  XMM_Error, bl
@Success:
end;


function XMM_GetHandleInfo( Handle : word;
                            var LockCount : byte;
                            var FreeEMB   : byte;
                            var Length    : word ) : boolean; assembler;
{ renvoi info sur l'handle }
asm
  mov  ah, 0Eh
  mov  dx, Handle
  call XMM_Control
  or   ax, ax
  jnz  @Success
  mov  XMM_Error, bl
  jmp  @Exit
@Success:
  les  di, LockCount
  mov  [es:di], bh
  les  di, FreeEMB
  mov  [es:di], bl
  les  di, Length
  mov  [es:di], dx
@Exit:
end;


function XMM_ReallocateExtended( Handle, NewSize : word ) : boolean; assembler;
{ modifie la taille d'une allocation }
asm
  mov  ah, 0Fh
  mov  dx, Handle
  mov  bx, NewSize
  call XMM_Control
  or   ax, ax
  jnz  @Success
  mov  XMM_Error, bl
@Success:
end;


function XMM_RequestUMB( Size : word;
                         var SegUMB : word;
                         var SizeUMB : word ) : boolean; assembler;
{ tente d'allouer un block en dessous des 1 Megas et au dessus des 640Ko
  Size est en paragraphe.
  Renvoi FALSE si erreur. si XMM_Result=B0h alors SizeUMB=taille max libre }
asm
  mov  ah, 10h
  mov  dx, Size
  call XMM_Control
  les  di, SizeUMB
  mov  [es:di], dx      { dans tous les cas, on note DX (SizeUMB) }
  or   ax, ax
  jnz  @Success
  mov  XMM_Error, bl
  jmp  @Exit
@Success:
  les  di, SegUMB
  mov  [es:di], bx
@Exit:
end;


function XMM_ReleaseUMB( SegUMB : word ) : boolean; assembler;
{ libre un block Upper Memory prcdement allou }
asm
  mov  ah, 11h
  mov  dx, SegUMB
  call XMM_Control
  or   ax, ax
  jnz  @Success
  mov  XMM_Error, bl
@Success:
end;


procedure AffErreur;
begin
  writeln('XMM Error : ',XMM_Result);
end;


function FoncTas( Taille : word ) : integer; far;
begin
  if Taille>0 then FoncTas := 1;
end;


function GetC18 : longint; assembler;
asm
  xor  ah, ah
  int  1Ah
  mov  ax, dx
  mov  dx, cx
end;

procedure SpeedTest;
const SizeBuf = 16384;
type Buffer = array[1..SizeBuf] of byte;
     PtrRec = record Ofs, Seg : word; end;
var Handle : word;
    LFree  : word;
    TFree  : word;
    Buf    : ^Buffer;
    EMM    : ExtMemMoveRec;
    j      : word;
    i      : word;
    t1, t2 : longint;
    p      : pointer;
    tt     : real;
    vol    : real;
    t      : real;
begin
  GetMem(p,sizeof(Buf^)+16);
  if p<>nil then
    begin
      if PtrRec(p).Ofs<>0 then Buf := Ptr(PtrRec(p).Seg+1,0)
      else Buf := Ptr(PtrRec(p).Seg,0);
      if XMM_QueryFreeExtended(LFree,TFree) then
        begin
          writeln('Largest Free: ',LFree:5,' Ko   Total Free: ',TFree:5,' Ko');
          if LFree>=SizeBuf div 1024 then
            begin
              if XMM_AllocateExtended(Handle,SizeBuf div 1024) then
                begin
                  for i := 1 to SizeBuf do Buf^[i] := byte(i); { mark }
                  writeln('Test de vitesse en cours...');

                  t1 := GetC18;
                  for j := 0 to 511 do
                    begin
                      FillChar(EMM,sizeof(EMM),0);
                      { envoi en mmoire }
                      EMM.Length       := SizeBuf;
                      EMM.SourceHandle := 0;
                      EMM.SourceOffset := longint(Buf);
                      EMM.DestHandle   := Handle;
                      EMM.DestOffset   := 0;
                      XMM_MoveExtended(EMM);

                      EMM.Length       := SizeBuf;
                      EMM.SourceHandle := Handle;
                      EMM.SourceOffset := 0;
                      EMM.DestHandle   := 0;
                      EMM.DestOffset   := longint(Buf);
                      XMM_MoveExtended(EMM);
                    end;
                  t2 := GetC18;
                  vol := (longint(j+1)*longint(SizeBuf)*longint(2))/(1024*1024);
                  t := (t2-t1)/18.2;
                  tt :=  (96/(vol*1024))*t;
                  writeln(vol:2:2,' mga swap en ',t:2:2,' s, l''idal pour ',
                          1/(8*tt):0:0,'  ',1/(4*tt):0:0,' tche(s) maximun.');

                  if not(XMM_FreeExtended(Handle)) then AffErreur;
                end
              else AffErreur;
            end
          else writeln('Pas assez de mmoire tendue (min:63 Ko)');
        end
      else AffErreur;
      FreeMem(p,sizeof(Buf^)+16);
    end
  else writeln('Pas assez de mmoire conventionnelle (min:63 Ko)');
end;


type LongRec = record
         Low, High : word;
       end;


Procedure UMBTest;
Var p : Word;
    r : Word;
Begin
 If XMM_RequestUMB($FFFF, p, r ) Then
  Begin
  End
 Else
  Begin
   AffErreur;
   Writeln('r=',r);
  End;
End;


Function GetUpperMemoryLink : Byte; Assembler;
Asm
    MOV  AX, 5802h
    INT  21h
    JNC  @FIN
    MOV  AL, 0FFh   { Erreur }
@FIN:
End;


Function SetUpperMemoryLink( Gu : Byte ) : Byte; Assembler;
Asm
    MOV  AX, 5802h
    INT  21h
    JC   @ERR            { Erreur }
    PUSH AX              { Sauve }
    MOV  AX, 5803h
    MOV  BL, [Gu]
    XOR  BH, BH
    INT  21h
    POP  AX              { Restore }
    JNC  @FIN
@ERR:
    MOV  AL, 0FFh
    JMP  @FIN
@FIN:
End;


Function GetAllocationStrategy : Word; Assembler;
Asm
    MOV  AX, 5800h
    INT  21h
End;


Function SetAllocationStrategy( as : Word ) : Word; Assembler;
Asm
    MOV  AX, 5800h
    INT  21h
    PUSH AX
    MOV  AX, 5801h
    MOV  BX, [as]
    INT  21h
    POP  AX
    JNC  @FIN
    MOV  AX, 0FFFFh
@FIN:
End;


Function DOSMemFree : Word; Assembler;
Asm
    MOV  AX, 4800h
    MOV  BX, 0FFFFh
    INT  21h
    MOV  AX, BX
End;


var ver : longint;
    gu  : Byte;
    as  : Word;
begin
  if XMM_Installed then
    begin

      ver := XMM_Version;
      writeln('XMM Version : ',Hi(LongRec(ver).Low),'.',
                               Lo(LongRec(ver).Low),'  (Internal:',
                               Hi(LongRec(ver).High),'.',
                               Lo(LongRec(ver).High),')');

      gu := SetUpperMemoryLink(1);

      Writeln('Ex Gu=',gu);

      as := SetAllocationStrategy( 0 );

      Writeln('as=',as);

      Writeln('MemFree=',DOSMemFree,' & ',MemAvail);

      UMBTest;

      SetAllocationStrategy( as );

      gu := SetUpperMemoryLink(gu);

      (*
      SpeedTest;
      *)
    end
  else
    writeln('Pas de driver XMS');
end.
